import asyncio

from asynconsumer.core import async_run
from . import to_md5


def fetch_gcs_objects(uris, directory, naming=None, concurrency=3, sleep=0):
    """GCSから取得したオブジェクトをdirectoryに取得し、そのファイルパスを返す。

    :param uris: uriの一覧
    :param directory: オブジェクトを格納するディレクトリのパス
    :param naming: uriを引数に取り、ファイル名を返すCallableオブジェクト。デフォルトはmd5化。
    :param concurrency: 並列実行の最大数。デフォルトは3。
    :param sleep: スリープ時間（秒）。デフォルトは0。
    :return: 渡したurisと同潤のファイルパスの一覧。取得できない場合はファイルパスはNoneにする。
    """
    from google.cloud import storage

    client = storage.Client()
    loop = asyncio.get_event_loop()
    naming = naming or to_md5

    async def _fetch(uri):
        local_path = directory + '/' + naming(uri)
        try:
            params = [client, uri, local_path]
            filename = await loop.run_in_executor(None, _get_gcs_object, *params)
        except Exception:
            filename = None
        return filename

    return async_run(uris, _fetch, concurrency=concurrency, sleep=sleep)


def _get_gcs_object(client, uri, filename):
    try:
        bucket_name, blob_name = uri[5:].split('/', 1)
        data = client.bucket(bucket_name).get_blob(blob_name)
    except Exception as e:
        raise e
    if not data:
        return None
    data.download_to_filename(filename)
    return filename



import asyncio

from asynconsumer.core import async_run


def nop(targets):
    async def coro(target):
        print('start: {}'.format(target))
        await asyncio.sleep(1)
        print('end: {}'.format(target))
        return target.upper()

    return async_run(targets, coro, concurrency=2)


def nop_2(targets):
    def heavy_synchronous_process(target):
        print('processing: {}'.format(target))
        return target.upper()

    loop = asyncio.get_event_loop()

    async def coro(target):
        params = [target]
        result = await loop.run_in_executor(None, heavy_synchronous_process, *params)
        return result

    return async_run(targets, coro, concurrency=2)



import asyncio
import itertools
import logging

logger = logging.getLogger(__name__)


async def _bound_coro(target,
                      coro,
                      semaphore,
                      sleep):
    """排他処理を行いながら非同期で処理を実行する

    :param urls: 処理対象一覧
    :param coro: 処理内容を記述したコルーチン
    :param semaphore: セマフォ。
    :param sleep: スリープ時間（秒）。
    :return: coroの戻り値
    """
    async with semaphore:
        await asyncio.sleep(sleep)
        return await coro(target)


async def _run(targets,
               coro,
               concurrency,
               sleep):
    """リストに対して非同期処理を行う

    :param urls: 処理対象一覧
    :param coro: 処理内容を記述したコルーチン
    :param concurrency: 並列実行の最大数。デフォルトは1。
    :param sleep: スリープ時間（秒）。
    :return: coroの戻り値
    """
    responses = []
    semaphore = asyncio.Semaphore(concurrency)
    for chunk in get_chunk(targets, n=concurrency * 10):
        tasks = (asyncio.ensure_future(_bound_coro(target, coro, semaphore, sleep))
                 for target in chunk)
        responses.extend(await asyncio.gather(*tasks))
        logger.info('{} targets are completed.'.format(len(responses)))
    return responses


def async_run(targets,
              coro,
              concurrency=1,
              sleep=0):
    """リストに対して非同期処理を行う

    :param urls: 処理対象一覧
    :param coro: 処理内容を記述したコルーチン
    :param concurrency: 並列実行の最大数。デフォルトは1。
    :param sleep: スリープ時間（秒）。
    :return: coroの戻り値の一覧。順番は保持される。
    """
    targets = targets or []
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(_run(targets, coro, concurrency, sleep))
    return results


def get_chunk(iterable, n=5):
    for i, item in itertools.groupby(enumerate(iterable), lambda x: x[0] // n):
        yield (x[1] for x in item)



import logging

from asynconsumer.core import async_run
from . import to_md5

logger = logging.getLogger(__name__)


def fetch_http_resources(urls, directory, naming=None, timeout=10, concurrency=3, sleep=0.5):
    """HTTPリソースをdirectoryに取得し、そのファイルパスを返す。

    :param urls: urlの一覧
    :param directory: オブジェクトを格納するディレクトリのパス
    :param naming: urlを引数に取り、ファイル名を返すCallableオブジェクト。デフォルトはmd5化。
    :param timeout: タイムアウト時間（秒）。
    :param concurrency: 並列実行の最大数。デフォルトは3。
    :param sleep: スリープ時間（秒）。デフォルトは0.5。
    :return: 渡したurisと同潤のファイルパスの一覧。取得できない場合はファイルパスはNoneにする。
    """
    import aiohttp
    import async_timeout

    naming = naming or to_md5

    async def _fetch(url):
        async with aiohttp.ClientSession() as session:
            async with async_timeout.timeout(timeout):
                try:
                    head = await session.head(url)
                    if head and 200 <= head.status < 300:
                        response = await session.get(url)
                        if response and not (200 <= response.status < 300):
                            if response.status != 404:
                                logger.info(
                                    'can not get some data: {} {}'.format(url, response.status))
                            response = None
                    else:
                        response = None
                except Exception as e:
                    logger.info('can not get some data: {} {}'.format(url, e))
                    response = None

            local_path = directory + '/' + naming(url)
            if response:
                with open(local_path, 'wb') as f:
                    f.write(await response.read())
            return local_path if response else None

    return async_run(urls, _fetch, concurrency=concurrency, sleep=sleep)



import hashlib


def to_byte(bytes_or_str, enc='utf-8'):
    """byteに変換して返す"""
    if isinstance(bytes_or_str, str):
        value = bytes_or_str.encode(enc)
    else:
        value = bytes_or_str or b''
    return value


def to_md5(data):
    return hashlib.md5(to_byte(data)).hexdigest()


from .gcs_downloader import fetch_gcs_objects
from .http_downloader import fetch_http_resources

__all__ = ['fetch_gcs_objects', 'fetch_http_resources']



from setuptools import setup, find_packages


def main():
    setup(
        name='asynconsumer',
        version='1.0.1',
        url='https://github.com/rhoboro/asynconsumer',
        license='Apache License 2.0',
        author='rhoboro',
        author_email='rhoboro@gmail.com',
        maintainer='rhoboro',
        maintainer_email='rhoboro@gmail.com',
        packages=find_packages(),
        install_requires=[],
        extras_require={
        },
        classifiers=[
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
    )


if __name__ == '__main__':
    main()
